import argparse
import asyncio
import logging

from aiojsonrpc2 import logger as jsonrpc_logger
import yaml

from . import app_version, logger as module_logger
from .errors import *
from .utils import import_from_module


logger = logging.getLogger(__name__)


class Proxy(object):
    def __init__(self, name='', **kwargs):
        self.name = name

        self.settings = kwargs
        self.pool_settings = kwargs.get('pools') or []
        if isinstance(self.pool_settings, dict):
            self.pool_settings = [self.pool_settings]

    async def startup(self):
        try:
            wklass = import_from_module(self.settings.get('worker_class') or '')
            pklass = import_from_module(self.settings.get('pool_class') or '')
        except (ModuleNotFoundError, AttributeError) as e:
            raise ConfigurationError(e)

        logger.info("* {} proxy starting".format(self.name))

        self.workers = wklass(self, self.settings.get('listen'), **self.settings)
        self.pool = pklass(self, self.pool_settings, **self.settings)

        await self.workers.initialize()
        await self.workers.start_listening()

        logger.info("* {} proxy started, waiting for worker connections".format(self.name))

    async def shutdown(self):
        logger.info("* {} proxy stopping".format(self.name))

        await self.workers.close()
        await self.pool.close()

        logger.info("* {} proxy stopped".format(self.name))


class Application(object):
    proxies = {}
    config = {}

    def __init__(self, config_file):
        self.config_file = config_file

    async def startup(self):
        try:
            with open(self.config_file, 'r') as cf:
                self.config = yaml.load(cf.read())
        except Exception:
            raise ConfigurationError("Unable to load configuration file")

        for n, settings in enumerate(self.config.get('proxies', []), 1):
            name = settings.pop('name', '') or 'Proxy {}'.format(n)
            proxy = Proxy(name=name, **settings)
            if proxy.name in self.proxies:
                raise ConfigurationError('A proxy named "{}" already exists; check config file'.format(proxy.name))

            try:
                await proxy.startup()
            except OSError as e:
                await proxy.shutdown()
                raise ServerAddressInUse(e)

            self.proxies[proxy.name] = proxy

    async def shutdown(self):
        await asyncio.gather(*[p.shutdown() for p in self.proxies.values()])
        self.proxies.clear()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-g", "--generate-config", action="store_true",
                        help="output a starting config file template")
    parser.add_argument("-c", "--config", dest="config_file",
                        help="path to configuration file")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="minimum output verbosity (>=WARNING)")
    parser.add_argument("-l", "--loud", action="store_true",
                        help="maximum output verbosity (>=DEBUG)")
    parser.add_argument("-v", "--version", action="version", version=app_version)
    args = parser.parse_args()

    if args.generate_config:
        from .utils import output_config
        output_config()
        return

    if args.loud:
        logf = logging.Formatter('%(asctime)s %(levelname)8s - %(message)s (%(name)s:%(lineno)d)')
    else:
        logf = logging.Formatter('%(asctime)s %(levelname)8s - %(message)s')

    logh = logging.StreamHandler()
    logh.setFormatter(logf)

    module_logger.addHandler(logh)
    jsonrpc_logger.addHandler(logh)
    module_logger.setLevel(logging.INFO)
    jsonrpc_logger.setLevel(logging.INFO)

    if args.loud:
        logger.info('* Verbose mode enabled')
        module_logger.setLevel(logging.DEBUG)
        jsonrpc_logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.info('* Quiet mode enabled')
        module_logger.setLevel(logging.WARNING)
        jsonrpc_logger.setLevel(logging.WARNING)

    app = Application(args.config_file)

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(app.startup())
        running = True
    except (ServerAddressInUse, ConfigurationError) as e:
        logger.critical(str(e))
        running = False

    if running:
        # Serve requests until Ctrl+C
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass

    loop.run_until_complete(app.shutdown())
    loop.close()



import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__appname__ = __package__.replace('_', '-')
__version__ = "1.1"

app_version = "{}/{}".format(__appname__, __version__)



import asyncio
import binascii
from collections import deque, OrderedDict
import logging
import socket
import struct

from aiojsonrpc2 import ServerProtocol, ClientProtocol

from ..errors import *

logger = logging.getLogger(__name__)


class BaseWorkerProtocol(ServerProtocol):
    pool = None
    pool_watchdog_fut = None

    registered_extra_nonce1_tails = set()

    # we'll optionally track the most recent n shares/solutions
    # for duplicate detection; this needs to be 'enabled' in
    # hook_validate_share_params by adding 'unique' data to be
    # checked (which likely differs by algo/coin)
    recent_shares = deque(maxlen=500)

    def __init__(self, proxy, connection_settings, *args, **kwargs):
        self.proxy = proxy
        self.settings = kwargs
        self.log_prefix = 'W:{}:'.format(self.proxy.name)

        super().__init__(connection_settings)

        mw = self.settings.get('max_workers')
        if mw is None:
            self.max_workers = 256
            logger.info("{} defaulting to {} max workers".format(self.log_prefix, mw, self.max_workers))
        else:
            try:
                self.max_workers = int(mw)
                if self.max_workers not in [1, 256, 65536]:
                    raise ValueError
            except (ValueError, TypeError):
                self.max_workers = 256
                logger.warning("{} invalid 'max_workers' setting ({}), defaulting to {} instead".format(self.log_prefix, mw, self.max_workers))

        if self.max_workers != 1:
            logger.info("{} up to {} workers supported (distinct nonce spaces)".format(self.log_prefix, self.max_workers))
        else:
            logger.info("{} solo worker mode (single nonce space)".format(self.log_prefix, self.max_workers))

    async def pool_watchdog(self):
        while not self.stopping:
            await asyncio.sleep(1)

            # only try to reconnect to the pool if we have existing client
            # connections
            if len(self.clients) and not self.pool.connected:
                while True:
                    try:
                        await self.pool.connect()
                        break
                    except:
                        if self.stopping:
                            return
                        await self.pool.use_next_pool_config()

                await self.pool.initialize()
                self.pool.set_ready()

    async def initialize(self):
        self.pool = self.proxy.pool
        self.pool_watchdog_fut = asyncio.ensure_future(self.pool_watchdog())

    async def loop(self, connection):
        _socket = connection.reader._transport.get_extra_info('socket')

        try:
            if hasattr(socket, 'SOL_SOCKET') and hasattr(socket, 'SO_KEEPALIVE'):
                _socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1) # Enable keepalive packets
            if hasattr(socket, 'SOL_TCP'):
                if hasattr(socket, 'TCP_KEEPIDLE'):
                    # Seconds before sending keepalive probes
                    _socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 60)
                if hasattr(socket, 'TCP_KEEPINTVL'):
                    # Interval in seconds between keepalive probes
                    _socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
                if hasattr(socket, 'TCP_KEEPCNT'):
                    # Failed keepalive probes before declaring other end dead
                    _socket.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 20)
        except:
            # Some socket features are not available on all platforms (Windows and macOS!)
            logger.exception("{} unable to set socket keep-alive due to platform constraints".format(self.log_prefix))

        if not self.pool.connected or not self.pool.is_ready():
            self.recent_shares.clear()

            # wait until the pool is subscribed, authorized, etc
            await self.pool.wait_until_ready()

        try:
            connection.extra['extra_nonce1_tail'] = self.get_extra_nonce1_tail()
        except MaxClientsConnected:
            await self.close_connection(connection)
            # connection.close()
            # self.client_connections.remove(connection)
            # self.cleanup_connection(connection)
            logger.warning("{} maximum number of {} workers reached, disconnecting".format(
                self.log_prefix, len(self.clients)))
            return

        connection.extra['subscriptions'] = {}

        await super().loop(connection)

    def cleanup_connection(self, connection):
        tail = connection.extra.get('extra_nonce1_tail')
        if tail:
            try:
                self.registered_extra_nonce1_tails.remove(tail)
            except:
                pass

    async def close(self):
        await super().close()
        await self.pool_watchdog_fut

    def get_extra_nonce1_tail(self):
        if self.max_workers != 1:
            if self.max_workers == 65536:
                # 2 bytes allows for 65536 (0000 through FFFF)
                _format = '>H'
            else:
                # 1 byte allows for 256 workers (00 through FF)
                _format = '>B'

            for i in range(0, self.max_workers):
                tail = binascii.hexlify(struct.pack(_format, i)).decode('ascii')

                if tail not in self.registered_extra_nonce1_tails:
                    self.registered_extra_nonce1_tails.add(tail)
                    return tail

            raise MaxClientsConnected


class BasePoolProtocol(ClientProtocol):
    workers = None

    pool_configs = []

    ready = asyncio.Event()

    subscriptions = {}
    extra_nonce1 = None
    extra_nonce2_size = None

    target_difficulty = None

    jobs = OrderedDict()
    current_job = None

    authorized_workers = {}
    unauthorized_workers = set()

    def __init__(self, proxy, connection_settings, *args, **kwargs):
        self.proxy = proxy
        self.settings = kwargs

        if isinstance(connection_settings, dict):
            self.pool_configs = [connection_settings]
        else:
            self.pool_configs = connection_settings

        self.log_prefix = 'P:{}:'.format(self.proxy.name)
        # start things up with the first pool configuration in the list!
        super().__init__(self.pool_configs.pop(0))

    async def use_next_pool_config(self):
        # If we get here, there was a pool disconnection and we should
        # try the next pool and/or exponentially back off our reconnection
        # attempts (local internet disconnection, perhaps???)

        # reset the ready indicator
        self.ready.clear()

        try:
            next_config = self.pool_configs.pop(0)
        except IndexError:
            # There wasn't another pool configuration available (no fallback pool!)
            next_config = None

            # No other pool to connect to, so let's wait a few seconds and
            # try to reconnect to the current and only pool settings
            logger.warning("{} waiting 10 seconds before reconnecting to current pool".format(self.log_prefix))
            await asyncio.sleep(10)

        if next_config:
            # Store the current (disconnected) config back in our pool config list
            self.pool_configs.append(self.connection_settings)
            # Switch to the next config!
            self.set_connection_config(next_config)

    async def initialize(self):
        self.workers = self.proxy.workers

    def is_ready(self):
        return self.ready.is_set()

    def set_ready(self):
        if not self.ready.is_set():
            self.ready.set()

    async def wait_until_ready(self):
        await self.ready.wait()

    def set_extra_nonce_data(self, extra_nonce1, extra_nonce2_size=None):
        self.extra_nonce1 = extra_nonce1
        self.extra_nonce2_size = extra_nonce2_size

    async def loop(self, connection):
        await super().loop(connection)

        if not self.stopping:
            # All client connections will need to be closed so they
            # auto-reconnect to resubscribe for the new nonce, etc
            await self.workers.close_all_connections()

            self.jobs.clear()
            self.current_job = None

            self.authorized_workers.clear()
            self.unauthorized_workers.clear()

            await self.use_next_pool_config()

    # async def _DELETE_THIS_test_method(self):
    #     await asyncio.sleep(4)
    #     logger.critical("{} kill connection for testing purposes".format(self.log_prefix))
    #     if self.connected:
    #         await self.connection.close()
    #         self.cleanup_connection(self.connection)
    #         self.connection = None
    #
    # # TODO DELETE this override
    # async def connect(self):
    #     await super().connect()
    #     asyncio.ensure_future(self._DELETE_THIS_test_method())



import asyncio
import logging

from .. import app_version
from ..errors import *
from . import BaseWorkerProtocol, BasePoolProtocol

logger = logging.getLogger(__name__)


class BaseStratumWorkerProtocol(BaseWorkerProtocol):
    async def hook_get_subscription_response_params(self, connection):
        extra_nonce1_tail = connection.extra.get('extra_nonce1_tail') or ''

        # `None` here because we don't need to support resuming subscriptions
        params = [None, self.pool.extra_nonce1 + extra_nonce1_tail]

        # if self.pool.extra_nonce2_size is `None`, it's a stratum-like
        # protocol that doesn't pass it around (ie. zcash & derivatives)
        if self.pool.extra_nonce2_size is not None:
            params.append(int(self.pool.extra_nonce2_size - len(extra_nonce1_tail) / 2))

        return params

    async def hook_validate_share_params(self, connection, params):
        logger.warning("{} hook_validate_share_params not implemented".format(self.log_prefix))
        # - Without an implementation in here, it's possible to have stale,
        # duplicate, or unauthorized shares submitted to the pool.
        # - Share submission differs enough between coins, making a default
        # implementation here problematic at best
        return params

    async def hook_post_subscribe(self, connection):
        logger.warning("{} hook_post_subscribe not implemented".format(self.log_prefix))
        # - This hook most likely needs to be implemented for most coin
        # protocol communication; sending mining.notify, etc to clients
        # - Coin protocol implementations differ enough to make this
        # problematic

    async def handle_mining_subscribe(self, connection, params, **kwargs):
        asyncio.ensure_future(self.hook_post_subscribe(connection))
        return await self.hook_get_subscription_response_params(connection)

    async def handle_mining_authorize(self, connection, params, **kwargs):
        if len(params) == 2:
            account_name, account_password = params
        elif len(params) == 1:
            account_name, = params
            account_password = ''
        else:
            raise JSONRPCInvalidParams

        # possible future auth ideas:
        # - proxy settings define the auth user/pass params for the pool connection
        #   - but enforce worker user/pass auth through local proxy settings (so not just anyone can join)
        #   - or allow any auth to proxy from miner/worker
        # - proxy settings allow the miner/worker to pass-thru user/pass to the pool; in this case,
        #   we'd store if the user was already authed
        #   - multiple miners can use the same user/pass OR use separate credentials

        return await self.pool.authorize(account_name, account_password)

    async def handle_mining_submit(self, connection, params, **kwargs):
        params = await self.hook_validate_share_params(connection, params)
        return await self.pool.submit(params)

    # async def handle_mining_extranonce_subscribe(self, connection, params, **kwargs):
    #     connection.extra['subscriptions']['mining.extranonce.subscribe'] = True
    #     return True
    #
    # async def handle_mining_suggest_difficulty(self, connection, params, **kwargs):
    #     pass
    #
    # async def handle_mining_suggest_target(self, connection, params, **kwargs):
    #     pass
    #
    # this one is also shown as unsupported in original stratum-mining-proxy?
    # async def handle_mining_get_transactions(self, connection, params, **kwargs):
    #     pass


class BaseStratumPoolProtocol(BasePoolProtocol):
    async def initialize(self):
        await super().initialize()

        await self.subscribe()
        await self.extranonce_subscribe()

    async def hook_subscription_request_params(self):
        # This is in its own method to allow subclassing for future coins.
        return []

    # async def hook_extra_nonce2_size(self):
    #     # This is in its own method to allow subclassing for future coins.
    #     # `extra_nonce2_size` isn't supported by some coins' mining.subscribe
    #     # (zcash, zclassic, etc), so calculate it differently by overriding
    #     # this method in a derived class
    #     return self.extra_nonce2_size

    async def hook_validate_job_params(self, params):
        logger.warning("{} hook_validate_job_params not implemented".format(self.log_prefix))

        # a good default is to return the job id and whether to clean old jobs
        # most stratum-based protocols seem to have job id first and clean_jobs last
        return params[0], params[-1]

    async def hook_set_target(self, params):
        try:
            self.target_difficulty = params[0]
        except IndexError:
            raise JSONRPCInvalidParams

    async def subscribe(self):
        response = await self.connection.rpc('mining.subscribe', await self.hook_subscription_request_params())
        if not response.success:
            logger.warning('{} mining.subscribe response error code {}, message "{}"'.format(
                self.log_prefix, response.data.get('code'), response.data.get('msg')))
            return False

        extra_nonce2_size = None
        subscriptions, extra_nonce1 = response.data[:2]
        if len(response.data) > 2:
            extra_nonce2_size = response.data[2]

        # if subscriptions is empty/None, then the pool doesn't support subscriptions
        if subscriptions:
            if isinstance(subscriptions, list):
                if isinstance(subscriptions[0], list):
                    # Some coins pass back a list of two-item subscription methods/ids
                    # pairs ie. [[m1, id1], [m2, id2]] includes mining.notify
                    for _method, _id in subscriptions:
                        self.subscriptions[_method] = _id
                else:
                    # Some coins pass back a list containing a single method/id
                    # pairs ie. [m1, id1] - it should be just mining.notify
                    self.subscriptions[subscriptions[0]] = subscriptions[1]
            else:
                # Some coins pass back just the subscription id for mining.notify
                # ie. id1 (like zcash, zclassic)
                self.subscriptions['mining.notify'] = subscriptions

        if extra_nonce1 != self.extra_nonce1:
            self.set_extra_nonce_data(extra_nonce1, extra_nonce2_size)
            # TODO: ensure all clients get updated nonces, etc - reset 'jobs'

        return True

    async def extranonce_subscribe(self):
        if self.settings.get('extranonce_subscribe', False):
            try:
                # use a timeout here because some pools (eg. MPOS??) don't reply to this
                # JSONRPC request (even though it's NOT a JSONRPC notification); 3 seconds
                # ought to be enough
                response = await self.connection.rpc('mining.extranonce.subscribe', timeout=5)
                return response.success and response.data
            except:
                logger.info("{} pool doesn't support 'mining.extranonce.subscribe'".format(self.log_prefix))
        return False

    def is_authorized(self, account_name, account_password):
        if account_name in self.authorized_workers:
            return self.authorized_workers.get(account_name) == account_password
        return False

    def get_auth_params(self, miner_account_name, miner_account_password):
        paccount_name = self.connection_settings.get('account_name', '')
        paccount_password = self.connection_settings.get('account_password', '')
        if not paccount_name:
            logger.error("{} no pool credentials (account name/password) are set".format(self.log_prefix))

        # If the pool's auth account doesn't specify a worker name, then
        # try to get and append one from the miner's params
        if '.' not in paccount_name:
            try:
                worker_name = miner_account_name.rsplit('.', 1)[1]
            except IndexError:
                worker_name = ''

            # concatenate the worker name to the pool's
            paccount_name = ".".join([s for s in [paccount_name, worker_name] if len(s)])

        return paccount_name, paccount_password

    async def authorize(self, account_name, account_password):
        paccount_name, paccount_password = self.get_auth_params(account_name, account_password)

        if self.is_authorized(paccount_name, paccount_password):
            return True

        result = False
        if paccount_name and paccount_name not in self.unauthorized_workers:
            response = await self.connection.rpc('mining.authorize', [paccount_name, paccount_password])
            if response.success:
                result = response.data
                if result:
                    self.authorized_workers[paccount_name] = paccount_password
                else:
                    logger.warning("{} pool authorization denied".format(self.log_prefix))
                    self.unauthorized_workers.add(paccount_name)

        return result

    async def submit(self, params):
        # params[0] is the account_name from the miner, 'translate'
        # it as necessary to the account name we need for the pool
        paccount_name, paccount_password = self.get_auth_params(params[0], '')

        if not self.is_authorized(paccount_name, paccount_password):
            raise JSONRPCUnauthorizedWorker

        params[0] = paccount_name

        logger.debug('{} mining.submit params sent to pool {}'.format(self.log_prefix, params))
        response = await self.connection.rpc('mining.submit', params)
        return response.success and response.data

    async def handle_mining_notify(self, connection, params, **kwargs):
        job_id, clean_jobs = await self.hook_validate_job_params(params)
        if job_id:
            if clean_jobs:
                # TODO: abandon/clean all current jobs
                self.jobs.clear()

            self.current_job = params
            self.jobs[job_id] = params

            # only keep the last 3 jobs around for handling older job share submits
            while len(self.jobs) > 3:
                # get the first job key-value pair and remove it from the jobs
                # as it will be an old job that shouldn't be worked on anymore
                k, v = next(iter(self.jobs.items()))
                self.jobs.pop(k)

            await self.workers.broadcast('mining.notify', params, is_notification=True)

    async def handle_mining_set_target(self, connection, params, **kwargs):
        await self.hook_set_target(params)
        await self.workers.broadcast('mining.set_target', params, is_notification=True)

    async def handle_mining_set_difficulty(self, connection, params, **kwargs):
        # TODO: add another hook for hook_set_difficulty if it
        # needs to be treated differently at the proxy level
        await self.hook_set_target(params)
        await self.workers.broadcast('mining.set_difficulty', params, is_notification=True)

    async def handle_client_get_version(self, connection, params, **kwargs):
        return app_version

    async def handle_client_show_message(self, connection, params, **kwargs):
        if len(params) != 1:
            raise JSONRPCInvalidParams
        await self.workers.broadcast('client.show_message', params)

    async def handle_mining_set_extranonce(self, connection, params, **kwargs):
        if len(params) != 2:
            raise JSONRPCInvalidParams

        self.set_extra_nonce_data(*params[:2])

        for conn in self.workers.clients.keys():
            # has the user subscribed to receive new extranonce notifications?
            if conn.extra.get('subscriptions', {}).get('mining.extranonce.subscribe'):
                tail = conn.extra.get('extra_nonce1_tail')
                if tail:
                    # copy the params...
                    _p = params[:]

                    # ... so we can add the connection's nonce 'tail' on (maintains
                    # distinct nonce spaces between multiple workers)
                    _p[0] = _p[0] + tail

                    # extra_nonce2_size is only set if the coin/algo supports it
                    # (eg. zcash/zclassic do not)
                    if self.extra_nonce2_size is not None:
                        _p[1] = int(self.extra_nonce2_size - len(tail) / 2)

                    await conn.rpc('mining.set_extranonce', _p, True)
            else:
                # Need to drop worker connections that aren't able (haven't subscribed!)
                # to receive new extranonce values? So they'll reconnect...
                self.workers.close_connection(conn)

    # async def handle_client_reconnect(self, connection, params, **kwargs):
    #     pass
    #
    # async def handle_client_add_peers(self, connection, params, **kwargs):
    #     pass
    #
    # async def handle_mining_get_hashrate(self, connection, params, **kwargs):
    #     pass
    #
    # async def handle_mining_get_temperature(self, connection, params, **kwargs):
    #     pass


class StratumWorkerProtocol(BaseStratumWorkerProtocol):
    # TODO: need to implement hooks; these are for bitcoin, litecoin, etc,  I believe?
    pass


class StratumPoolProtocol(BaseStratumPoolProtocol):
    # TODO: need to implement hooks; these are for bitcoin, litecoin, etc,  I believe?
    pass



import logging

from .. import app_version
from ..errors import *
from .stratum import BaseStratumPoolProtocol, BaseStratumWorkerProtocol

logger = logging.getLogger(__name__)


class EquihashWorkerProtocol(BaseStratumWorkerProtocol):
    async def hook_post_subscribe(self, connection):
        # checks around these to ensure the first miner connecting doesn't get
        # sent these notification before the pool sends this proxy the initial
        # values for them! (otherwise, we send junk values)
        if self.pool.target_difficulty is not None:
            await connection.rpc('mining.set_target', [self.pool.target_difficulty], is_notification=True)
        if self.pool.current_job is not None:
            await connection.rpc('mining.notify', self.pool.current_job, is_notification=True)

    async def hook_validate_share_params(self, connection, params):
        if len(params) == 5:
            # account_name, job_id, time, nonce2, equihash_solution
            job_id = params[1]

            # handle the distinct nonce spacing by prepending the nonce1
            # tail to the nonce2 from the worker
            nonce2 = connection.extra['extra_nonce1_tail'] + params[-2]
            params[-2] = nonce2

            if job_id not in self.pool.jobs:
                raise JSONRPCJobNotFound

            check = (job_id, nonce2)
            if check in self.recent_shares:
                raise JSONRPCDuplicateShare

            self.recent_shares.append(check)

            return params

        raise JSONRPCInvalidParams


class EquihashPoolProtocol(BaseStratumPoolProtocol):
    async def hook_subscription_request_params(self):
        return [
            app_version,
            self.subscriptions.get('mining.notify'),
            self.connection_settings.get('host'),
            self.connection_settings.get('port')
        ]

    # async def hook_extra_nonce2_size(self):
    #     return int(32 - len(self.extra_nonce1) / 2)

    async def hook_validate_job_params(self, params):
        # normally only 8 params, some pools send another bool at the end
        if len(params) in (8, 9):
            # job_id, version, prevhash, merkleroot, reserved, time, bits, clean_jobs
            job_id, job_version = params[:2]

            # 04000000: zcash, other equihash derivatives;
            # 00000020: bitcoin gold, zencash
            if job_version in ["04000000", "00000020"]:
                clean_jobs = params[7]
                return job_id, clean_jobs

        raise JSONRPCInvalidParams




from datetime import datetime, timezone
from importlib import import_module

from . import app_version


def import_from_module(s):
    p, m = s.rsplit('.', 1)
    mod = import_module(p)
    return getattr(mod, m)


default_config = """# This file was generated by {app_version} on {generated_datetime}

proxies:
- ## Optional name to give your log output some flair and order

  #name: ''

  ## max_workers allows the proxy to ensure each worker up to this limit
  ## has a unique nonce-space in which to generate solutions, avoiding
  ## duplicate shares from multiple rigs.
  ## Options are 1 (solo mode), 256 (the default), or 65536

  #max_workers: 256

  ## Some pools for some algorithms support sending out a new extra_nonce1
  ## value to connected workers; some pools do not support this, and will
  ## disconnect immediately if this is set to true

  #extranonce_subscribe: false

  ## These two lines define the aiostratum_proxy Python classes you
  ## want to use to handle this proxy's workers and pool connections

  worker_class: aiostratum_proxy.protocols.equihash.EquihashWorkerProtocol
  pool_class: aiostratum_proxy.protocols.equihash.EquihashPoolProtocol

  ## This is a list of host/port pairs to which your workers can connect
  ## to this proxy; blank `host` (the default) means listen on all available
  ## interfaces (all private/public IPs; both IPv4/6)

  listen:
  - host: ''
    port: 10666
  # - host: ''
  #   port: 10667
  #   ## TLS/SSL supported; add another port listening for TLS/SSL connections
  #   ssl: true
  #   ssl_cert_file: '<path to your ssl cert file>'
  #   ssl_cert_key_file: '<path to your ssl cert key file>'

  ## This is the list of pools (at least 1 required, obviously) to have the
  ## proxy connect to

  pools:
  - name: Example primary pool
    host: btcp.pooldomain.io
    port: 9000
    account_name: poolaccountnameoraddress1.rig1
    account_password: poolaccountpassword
  # - name: Example fallback pool #1
  #   host: btcp.anotherpooldomain.io
  #   port: 9001
  #   ## Miner login credentials for pool
  #   account_name: poolaccountnameoraddress2.rig1
  #   account_password: poolaccountpassword
  #   ## TLS/SSL pool connection? (check with your pool operator if supported),
  #   ## and should the SSL connection verify the server's certificate?
  #   ssl: true
  #   ssl_verify: false
""".format(
    app_version=app_version,
    generated_datetime=datetime.now(timezone.utc).astimezone().isoformat()
)


def output_config():
    print(default_config)



from setuptools import setup, find_packages
from os import path

import aiostratum_proxy

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    from codecs import open
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='aiostratum_proxy',
    version=aiostratum_proxy.__version__,
    description='Python3 asyncio stratum protocol mining proxy',
    long_description=long_description,
    keywords='async asyncio aio crypto cryptocurrency blockchain '
             'mining stratum protocol proxy pool equihash bitcoin'
             'litecoin',

    license='MIT',
    author='wetblanketcc',
    author_email='35851045+wetblanketcc@users.noreply.github.com',
    url='https://github.com/wetblanketcc/aiostratum_proxy',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: AsyncIO',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Networking',
    ],

    packages=find_packages(),

    python_requires='>=3.5',
    install_requires=[
        'aiojsonrpc2==1.0.0',
        'PyYAML==3.12',
    ],

    entry_points = {
        'console_scripts': [
            'aiostratum-proxy=aiostratum_proxy.application:main'
        ],
    }
)



from aiojsonrpc2.errors import *


class AIOStratumError(Exception):
    pass


class ConfigurationError(AIOStratumError):
    pass


class NetworkError(AIOStratumError):
    pass


class ServerAddressInUse(NetworkError):
    pass


class MaxClientsConnected(AIOStratumError):
    pass


class JSONRPCOtherUnknownError(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(20, msg or "Other/Unknown")


class JSONRPCJobNotFound(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(21, msg or "Job not found (=stale)")


class JSONRPCDuplicateShare(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(22, msg or "Duplicate share")


class JSONRPCLowDifficultyShare(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(23, msg or "Low difficulty share")


class JSONRPCUnauthorizedWorker(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(24, msg or "Unauthorized worker")


class JSONRPCNotSubscribed(JSONRPCError):
    def __init__(self, msg=''):
        super().__init__(25, msg or "Not subscribed")
